#include <GL/glew.h>

#include "mesh-gl.h"

MeshGL::MeshGL()
{
	which_mask = 0;

	vao = 0;

	vbo_vertices = 0;
	vbo_normals = 0;
	ibo = 0;

	num_vertices = 0;
	num_indices = 0;
}

MeshGL::MeshGL(const MeshGL &mesh)
{
	indices = mesh.indices;
	vertices = mesh.vertices;
	normals = mesh.normals;

	vbo_vertices = 0;
	vbo_normals = 0;
	ibo = 0;

	/*
	 * if we set these to the actual
	 * vertices.size() and indices.size()
	 * update_vbo will have no effect
	 * */

	num_vertices = 0;
	num_indices = 0;
}

MeshGL &MeshGL::operator=(const MeshGL &mesh)
{
	if(this == &mesh)
		return *this;

	/* to avoid OpenGL leaks */
	destroy_vbo();

	/* what the copy constructor does */
	indices = mesh.indices;
	vertices = mesh.vertices;
	normals = mesh.normals;

	return *this;
}

MeshGL::~MeshGL()
{
	destroy_vbo();

	vertices.clear();
	normals.clear();
}

/*
void MeshGL::draw() const
{
	// save the previously bound vao, vbo just in case
	int curr_vao;
	int curr_vbo;
	glGetIntegerv(GL_VERTEX_ARRAY_BINDING, &curr_vao);
	glGetIntegerv(GL_ELEMENT_ARRAY_BUFFER_BINDING, &curr_vbo);

	glBindVertexArray(vao);

	glBindBuffer(GL_ARRAY_BUFFER, vbo_vertices);
	glVertexPointer(3, GL_FLOAT, 0, 0);
	glBindBuffer(GL_ARRAY_BUFFER, vbo_normals);
	glNormalPointer(GL_FLOAT, 0, 0);
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ibo);
	glDrawElements(GL_TRIANGLES, indices.size(), GL_UNSIGNED_SHORT, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);

	// return to previous state
	glBindBuffer(GL_ARRAY_BUFFER, curr_vbo);
	glBindVertexArray(curr_vao);
}
*/

void MeshGL::update_vertex_data()
{
	update_vbo();
}

void MeshGL::update_vbo()
{
	// save the previously bound vao, vbo just in case
	int curr_vao;
	int curr_vbo;
	glGetIntegerv(GL_VERTEX_ARRAY_BINDING, &curr_vao);
	glGetIntegerv(GL_ELEMENT_ARRAY_BUFFER_BINDING, &curr_vbo);

	if(!vao)
		glGenVertexArrays(1, &vao);

	glBindVertexArray(vao);

	if(which_mask & MESH_NORMAL) {
		if(!vbo_normals) {
			glGenBuffers(1, &vbo_normals);
		}
		glBindBuffer(GL_ARRAY_BUFFER, vbo_normals);
		if(num_vertices != (int)normals.size()) {
			glBufferData(GL_ARRAY_BUFFER, normals.size() * 3 * sizeof(float),
			             &normals[0], GL_STREAM_DRAW);
		}
		else {
			glBufferSubData(GL_ARRAY_BUFFER, 0, normals.size() * 3 * sizeof(float),
			                &normals[0]);
		}
	}

	if(which_mask & MESH_VERTEX) {
		if(!vbo_vertices) {
			glGenBuffers(1, &vbo_vertices);
		}
		glBindBuffer(GL_ARRAY_BUFFER, vbo_vertices);
		if(num_vertices != (int)vertices.size()) {
			glBufferData(GL_ARRAY_BUFFER, vertices.size() * 3 * sizeof(float),
			             &vertices[0], GL_STREAM_DRAW);
		}
		else {
			glBufferSubData(GL_ARRAY_BUFFER, 0, vertices.size() * 3 * sizeof(float),
			                &vertices[0]);
		}
		num_vertices = vertices.size();
	}

	if(which_mask & MESH_INDEX) {
		if(!ibo) {
			glGenBuffers(1, &ibo);
		}
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ibo);
		if(num_indices != (int)indices.size()) {
			glBufferData(GL_ELEMENT_ARRAY_BUFFER, indices.size() * 2,
			             &indices[0], GL_STATIC_DRAW);
		}
		else {
			glBufferSubData(GL_ELEMENT_ARRAY_BUFFER, 0, indices.size() * 2,
			                &indices[0]);
		}
		num_indices = indices.size();
	}

	/* bind previously bound vbo */
	glBindBuffer(GL_ARRAY_BUFFER, curr_vbo);
	glBindVertexArray(curr_vao);
}

void MeshGL::destroy_vbo()
{
	if(vbo_vertices)
		glDeleteBuffers(1, &vbo_vertices);
	if(vbo_normals)
		glDeleteBuffers(1, &vbo_normals);
	if(ibo)
		glDeleteBuffers(1, &ibo);
	if(vao)
		glDeleteVertexArrays(1, &vao);
}